import os
import datetime
import shutil
import zipfile
import logging
import boto3
from boto3.s3.transfer import S3Transfer, TransferConfig
from botocore.exceptions import BotoCoreError

from ..flumine import FLUMINE_DATA

logger = logging.getLogger(__name__)


class BaseEngine:

    NAME = None

    def __init__(self, directory, market_expiration=3600, force_update=True):
        self.directory = directory
        self.market_expiration = market_expiration
        self.force_update = force_update
        self.markets_loaded = []
        self.validate_settings()

    def __call__(self, market_id, market_definition, stream_id):
        # check that market has not already been loaded
        if market_id in self.markets_loaded:
            if self.force_update:
                logger.warning(
                    "File: /%s/%s/%s has already been loaded, updating.."
                    % (FLUMINE_DATA, stream_id, market_id)
                )
            else:
                logger.warning(
                    "File: /%s/%s/%s has already been loaded, NOT updating.."
                    % (FLUMINE_DATA, stream_id, market_id)
                )
                return

        logger.info("Loading %s to %s:%s" % (market_id, self.NAME, self.directory))
        file_dir = os.path.join(FLUMINE_DATA, stream_id, market_id)

        # check that file actually exists
        if not os.path.isfile(file_dir):
            logger.error(
                "File: %s does not exist in /%s/%s/"
                % (FLUMINE_DATA, market_id, stream_id)
            )
            return

        # zip file
        zip_file_dir = self.zip_file(file_dir, market_id, stream_id)

        # core load code
        self.load(zip_file_dir, market_definition)

        # clean up
        self.clean_up(stream_id)

        self.markets_loaded.append(market_id)

    def load(self, zip_file_dir, market_definition):
        """Loads zip_file_dir to expected dir
        """
        raise NotImplementedError

    def validate_settings(self):
        """Validates settings e.g. directory
        provided, raises OSError
        """
        pass

    @staticmethod
    def create_metadata(market_definition):
        try:
            del market_definition["runners"]
        except KeyError:
            pass
        return dict([a, str(x)] for a, x in market_definition.items())

    def clean_up(self, stream_id):
        """If zip > market_expiration old remove
        zip and txt file
        """
        directory = os.path.join(FLUMINE_DATA, stream_id)

        for file in os.listdir(directory):
            if file.endswith(".zip"):
                file_stats = os.stat(os.path.join(directory, file))
                last_modified = datetime.datetime.fromtimestamp(file_stats.st_mtime)
                seconds_since = (
                    datetime.datetime.now() - last_modified
                ).total_seconds()

                if seconds_since > self.market_expiration:
                    logger.info(
                        "Removing: %s, age: %ss" % (file, round(seconds_since, 2))
                    )

                    txt_path = os.path.join(directory, file.split(".zip")[0])
                    zip_path = os.path.join(directory, file)

                    os.remove(txt_path)
                    os.remove(zip_path)

    @staticmethod
    def zip_file(file_dir, market_id, stream_id):
        """zips txt file into filename.zip
        """
        zip_file_directory = os.path.join(FLUMINE_DATA, stream_id, "%s.zip" % market_id)
        with zipfile.ZipFile(zip_file_directory, mode="w") as zf:
            zf.write(
                file_dir, os.path.basename(file_dir), compress_type=zipfile.ZIP_DEFLATED
            )
        return zip_file_directory

    @property
    def markets_loaded_count(self):
        return len(self.markets_loaded)

    @property
    def extra(self):
        return {
            "name": self.NAME,
            "markets_loaded": self.markets_loaded,
            "markets_loaded_count": self.markets_loaded_count,
            "directory": self.directory,
        }


class Local(BaseEngine):

    NAME = "LOCAL"

    def validate_settings(self):
        if not os.path.isdir(self.directory):
            raise OSError("File dir %s does not exist" % self.directory)

    def load(self, zip_file_dir, market_definition):
        shutil.copy(zip_file_dir, self.directory)


class S3(BaseEngine):

    NAME = "S3"

    def __init__(
        self,
        directory,
        access_key=None,
        secret_key=None,
        data_type="marketdata",
        **kwargs
    ):
        self.s3 = boto3.client(
            "s3", aws_access_key_id=access_key, aws_secret_access_key=secret_key
        )
        transfer_config = TransferConfig(use_threads=False)
        self.transfer = S3Transfer(self.s3, config=transfer_config)
        self.data_type = data_type
        super(S3, self).__init__(directory, **kwargs)

    def validate_settings(self):
        # error raised if bucket not present
        self.s3.head_bucket(Bucket=self.directory)

    def load(self, zip_file_dir, market_definition):
        event_type_id = (
            market_definition.get("eventTypeId", 0) if market_definition else "7"
        )
        try:
            self.transfer.upload_file(
                filename=zip_file_dir,
                bucket=self.directory,
                key=os.path.join(
                    self.data_type,
                    "streaming",
                    event_type_id,
                    os.path.basename(zip_file_dir),
                ),
                extra_args={
                    "Metadata": self.create_metadata(market_definition)
                    if market_definition
                    else {}
                },
            )
            logger.info("%s successfully loaded to s3" % zip_file_dir)
        except (BotoCoreError, Exception) as e:
            logger.error("Error loading to s3: %s" % e)



import os
import logging
import threading
from tenacity import retry, wait_exponential
from betfairlightweight import APIClient, BetfairError

from .listener import FlumineListener
from .exceptions import RunError, StreamError

logger = logging.getLogger(__name__)

# dir used to store all data
FLUMINE_DATA = "/data"


class Flumine:
    def __init__(self, recorder, settings=None, unique_id=1e3):
        self._certificate_login = (
            settings.pop("certificate_login", True) if settings else True
        )
        self.trading = self._create_client(settings)
        self.recorder = recorder
        self.unique_id = unique_id

        self._running = False
        self._socket = None
        self.listener = FlumineListener(recorder)  # output queue hack

    def start(
        self,
        heartbeat_ms=None,
        conflate_ms=None,
        segmentation_enabled=None,
        async_=True,
    ):
        """Checks trading is logged in, creates socket,
        subscribes to markets, sets running to True and
        starts handler/run threads.
        """
        logger.info("Starting stream: %s" % self.unique_id)
        if self._running:
            raise RunError("Flumine is already running, call .stop() first")
        if async_:
            threading.Thread(
                target=self._run,
                args=(conflate_ms, heartbeat_ms, segmentation_enabled),
                daemon=True,
            ).start()
        else:
            self._run(conflate_ms, heartbeat_ms, segmentation_enabled)

    def stop(self):
        """Stops socket, sets running to false
        and socket to None
        """
        logger.info("Stopping stream: %s" % self.unique_id)
        if not self._running:
            raise RunError("Flumine is not running")
        if self._socket:
            self._socket.stop()
        self._running = False
        self._socket = None

    def stream_status(self):
        """Checks sockets status
        """
        return str(self._socket) if self._socket else "Socket not created"

    @staticmethod
    def _create_client(settings):
        """Returns APIClient based on settings
        or looks for username in os.environ
        """
        if settings:
            return APIClient(**settings.get("betfairlightweight"))
        else:
            username = os.environ.get("username")
            return APIClient(username=username)

    @retry(wait=wait_exponential(multiplier=1, min=2, max=20))
    def _run(self, conflate_ms, heartbeat_ms, segmentation_enabled):
        """ Runs socket and catches any errors, will
        attempt reconnect after 2s exponentially backing
        off to 20s.
        """
        self._check_login(force=True)

        self._create_socket()

        if self.recorder.STREAM_TYPE == "market":
            logger.info("Subscribing to markets")
            try:
                self.unique_id = self._socket.subscribe_to_markets(
                    market_filter=self.recorder.market_filter,
                    market_data_filter=self.recorder.market_data_filter,
                    conflate_ms=conflate_ms,
                    heartbeat_ms=heartbeat_ms,
                    segmentation_enabled=segmentation_enabled,
                    initial_clk=self.listener.initial_clk,
                    clk=self.listener.clk,
                )
            except BetfairError as e:
                logger.error("Betfair subscribe_to_markets error: %s" % e)
                raise
        elif self.recorder.STREAM_TYPE == "race":
            logger.info("Subscribing to races")
            try:
                self.unique_id = self._socket.subscribe_to_races()
            except BetfairError as e:
                logger.error("Betfair subscribe_to_races error: %s" % e)
                raise
        else:
            raise StreamError(
                "%s is not a valid stream type" % self.recorder.STREAM_TYPE
            )

        self._running = True
        try:
            logger.info("Starting socket..")
            self._socket.start(async_=False)
        except BetfairError as e:
            logger.error("Betfair error: %s" % e)
            raise
        except Exception as e:
            logger.error("Unknown error: %s" % e)
            raise

    def _check_login(self, force=False):
        """Login if session expired
        """
        if self.trading.session_expired or force:
            if self._certificate_login:
                try:
                    self.trading.login()
                except BetfairError as e:
                    logger.error("Betfair login error: %s" % e)
                    raise
            else:
                try:
                    self.trading.login_interactive()
                except BetfairError as e:
                    logger.error("Betfair login_interactive error: %s" % e)
                    raise

    def _create_socket(self):
        """Creates stream
        """
        logger.info("Creating socket")
        self._socket = self.trading.streaming.create_stream(
            unique_id=self.unique_id,
            description="Flumine Socket",
            listener=self.listener,
            host=self.recorder.HOST,
        )

    @property
    def running(self):
        return True if self._running else False

    @property
    def status(self):
        return "running" if self._running else "not running"

    def __repr__(self):
        return "<Flumine>"

    def __str__(self):
        return "<Flumine [%s]>" % self.status



import logging

from .flumine import Flumine
from . import resources
from .exceptions import FlumineException


__title__ = "flumine"
__version__ = "0.8.0"
__author__ = "Liam Pauling"


# Set default logging handler to avoid "No handler found" warnings.
class NullHandler(logging.Handler):
    def emit(self, record):
        pass


logging.getLogger(__name__).addHandler(NullHandler())



import logging
from betfairlightweight.streaming.listener import StreamListener
from betfairlightweight.streaming.stream import BaseStream

from .exceptions import ListenerError


"""
Custom listener that doesn't do any processing,
helps reduce CPU. Hacks the output queue with a
recorder to remove the need for a handler thread.
"""

logger = logging.getLogger(__name__)


class FlumineListener(StreamListener):
    def __init__(self, recorder, max_latency=0.5):
        super(FlumineListener, self).__init__(recorder, max_latency)
        self.recorder = recorder

    def _add_stream(self, unique_id, stream_type):
        if stream_type == "marketSubscription":
            return FlumineStream(self)
        elif stream_type == "orderSubscription":
            raise ListenerError("Not expecting an order stream...")
        elif stream_type == "raceSubscription":
            return FlumineRaceStream(self)


class FlumineStream(BaseStream):

    _lookup = "mc"

    def _process(self, market_books, publish_time):
        for market_book in market_books:
            market_id = market_book.get("id")
            if (
                "marketDefinition" in market_book
                and market_book["marketDefinition"]["status"] == "CLOSED"
            ):
                if market_id in self._caches:
                    # removes closed market from cache
                    del self._caches[market_id]
                    logger.info(
                        "[MarketStream: %s] %s removed, %s markets in cache"
                        % (self.unique_id, market_id, len(self._caches))
                    )
            elif self._caches.get(market_id) is None:
                # adds empty object to cache to track live market count
                self._caches[market_id] = object()
                logger.info(
                    "[MarketStream: %s] %s added, %s markets in cache"
                    % (self.unique_id, market_id, len(self._caches))
                )

        self.output_queue(market_books, publish_time)
        self._updates_processed += len(market_books)

    def __str__(self):
        return "FlumineStream"

    def __repr__(self):
        return "<FlumineStream [%s]>" % len(self._caches)


class FlumineRaceStream(BaseStream):

    _lookup = "rc"

    def _process(self, race_updates, publish_time):
        for update in race_updates:
            market_id = update["mid"]
            if self._caches.get(market_id) is None:
                # adds empty object to cache to track live market count
                self._caches[market_id] = object()
                logger.info(
                    "[RaceStream: %s] %s added, %s markets in cache"
                    % (self.unique_id, market_id, len(self._caches))
                )

        self.output_queue(race_updates, publish_time)
        self._updates_processed += len(race_updates)

    def __str__(self):
        return "FlumineRaceStream"

    def __repr__(self):
        return "<FlumineRaceStream [%s]>" % len(self._caches)



import os
import json
import logging
from betfairlightweight.filters import (
    streaming_market_filter,
    streaming_market_data_filter,
)

from ..utils import create_short_uuid
from ..flumine import FLUMINE_DATA

logger = logging.getLogger(__name__)


class BaseRecorder:
    """Base recorder which connects to all
    markets by default.
    """

    NAME = "BASE_RECORDER"
    STREAM_TYPE = None
    MARKET_ID_LOOKUP = None
    HOST = None

    def __init__(self, storage_engine, market_filter=None, market_data_filter=None):
        self.storage_engine = storage_engine
        self.market_filter = market_filter or streaming_market_filter()
        self.market_data_filter = market_data_filter or streaming_market_data_filter(
            fields=[
                "EX_ALL_OFFERS",
                "EX_TRADED",
                "EX_TRADED_VOL",
                "EX_LTP",
                "EX_MARKET_DEF",
                "SP_TRADED",
                "SP_PROJECTED",
            ]
        )
        self.stream_id = (
            create_short_uuid()
        )  # used to differentiate markets /<FLUMINE_DATA>/<stream_id>
        self.live_markets = []  # list of markets to be processed
        self._setup()
        logger.info("Recorder created %s" % self.stream_id)

    def __call__(self, updates, publish_time):
        """Checks market using market book parameters
        function then passes market_book to be processed.

        :param updates: List of updates
        :param publish_time: Publish time of market book
        """
        for update in updates:
            market_id = update.get(self.MARKET_ID_LOOKUP)
            self.check_market_book(market_id, update)
            if market_id in self.live_markets:
                self.process_update(update, publish_time)

    def check_market_book(self, market_id, market_book):
        """Logic used to decide if market_book should
        be processed

        :param market_id: Market id
        :param market_book: Market Book object
        """
        if market_id not in self.live_markets:
            self.live_markets.append(market_id)

    def process_update(self, update, publish_time):
        """Function that processes update

        :param update: update object
        :param publish_time: Publish time of market book
        """
        raise NotImplementedError

    def on_market_closed(self, update):
        """Function run when market is closed, this
        may execute more than once if update received
        after being closed.
        """
        market_id = update.get(self.MARKET_ID_LOOKUP)
        market_definition = update.get("marketDefinition")
        logger.info("Closing market %s" % market_id)
        self.storage_engine(market_id, market_definition, self.stream_id)

    def _setup(self):
        """Create stream folder in <FLUMINE_DATA>
        """
        directory = os.path.join(FLUMINE_DATA, self.stream_id)
        if not os.path.exists(directory):
            os.makedirs(directory)

    def __str__(self):
        return "<%s>" % self.NAME


class MarketRecorder(BaseRecorder):
    """Data recorder, records stream data
    to /<FLUMINE_DATA>/<stream_id>/market_id,
    a single market per file.
    """

    NAME = "MARKET_RECORDER"
    STREAM_TYPE = "market"
    MARKET_ID_LOOKUP = "id"

    def process_update(self, market_book, publish_time):
        filename = "%s" % market_book.get(self.MARKET_ID_LOOKUP)
        file_directory = os.path.join(FLUMINE_DATA, self.stream_id, filename)

        with open(file_directory, "a") as outfile:
            outfile.write(
                json.dumps(
                    {"op": "mcm", "clk": None, "pt": publish_time, "mc": [market_book]}
                )
                + "\n"
            )

        if (
            "marketDefinition" in market_book
            and market_book["marketDefinition"]["status"] == "CLOSED"
        ):
            self.on_market_closed(market_book)


class RaceRecorder(BaseRecorder):

    NAME = "RACE_RECORDER"
    STREAM_TYPE = "race"
    MARKET_ID_LOOKUP = "mid"
    HOST = "race"

    def process_update(self, update, publish_time):
        filename = "%s" % update.get(self.MARKET_ID_LOOKUP)
        file_directory = os.path.join(FLUMINE_DATA, self.stream_id, filename)

        with open(file_directory, "a") as outfile:
            outfile.write(
                json.dumps(
                    {"op": "rcm", "clk": None, "pt": publish_time, "rc": [update]}
                )
                + "\n"
            )

        # todo validate this is correct
        if "rpc" in update:
            rpc = update["rpc"]
            if (
                rpc["g"] == "Finish"
                and rpc["prg"] == 0
                and rpc["ord"] == []
                and "rrc" not in update
            ):
                self.on_market_closed(update)



class FlumineException(Exception):
    """Base class for Flumine Errors"""

    pass


class RunError(FlumineException):
    """Error raised if start() called when already running"""

    def __init__(self, message):
        super(RunError, self).__init__(message)


class ListenerError(FlumineException):
    """Error raised if error in Listener"""

    def __int__(self, message):
        super(ListenerError, self).__init__(message)


class StreamError(FlumineException):
    """Error raise if invalid stream_type provided"""

    def __int__(self, message):
        super(StreamError, self).__init__(message)



import os
import re
from setuptools import setup


with open(os.path.join(os.path.dirname(__file__), "requirements.txt")) as f:
    INSTALL_REQUIRES = f.read().splitlines()

with open("flumine/__init__.py", "r") as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE
    ).group(1)

setup(
    name="flumine",
    version=version,
    packages=["flumine", "flumine.resources", "flumine.storage"],
    package_dir={"flumine": "flumine"},
    install_requires=INSTALL_REQUIRES,
    url="https://github.com/liampauling/flumine",
    license="MIT",
    author="liampauling",
    author_email="",
    description="Betfair data record framework utilising betfair streaming",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    test_suite="tests",
)
